@@ -9,6 +9,7 @@ import tempfile
 import warnings
 from typing import Iterable
 from cad_to_dagmc import __version__
+from .direct_mesh_plugin import to_mesh
 
 
 def define_moab_core_and_tags() -> tuple[core.Core, dict]:
@@ -105,6 +106,9 @@ def vertices_to_h5m(
 
     moab_core, tags = define_moab_core_and_tags()
 
+    # Add the vertices once at the start
+    all_moab_verts = moab_core.create_vertices(vertices)
+
     volume_sets_by_solid_id = {}
     for material_tag, (solid_id, triangles_on_each_face) in zip(
         material_tags, triangles_by_solid_by_face.items()
@@ -144,14 +148,21 @@ def vertices_to_h5m(
 
                 moab_core.tag_set_data(tags["surf_sense"], face_set, sense_data)
 
-                moab_verts = moab_core.create_vertices(vertices)
+                # Collect only the vertices that lie on triangles on this face
+                face_vertices_set = set()
+                for triangle in triangles_on_face:
+                    face_vertices_set.update(triangle)
+                face_vertices_list = sorted(face_vertices_set)
+
+                # Only add these to the MOAB face
+                moab_verts = [all_moab_verts[ii] for ii in face_vertices_list]
                 moab_core.add_entity(face_set, moab_verts)
 
                 for triangle in triangles_on_face:
                     tri = (
-                        moab_verts[int(triangle[0])],
-                        moab_verts[int(triangle[1])],
-                        moab_verts[int(triangle[2])],
+                        all_moab_verts[int(triangle[0])],
+                        all_moab_verts[int(triangle[1])],
+                        all_moab_verts[int(triangle[2])],
                     )
 
                     moab_triangle = moab_core.create_element(types.MBTRI, tri)
@@ -579,7 +590,7 @@ class CadToDagmc:
         cadquery_object: (
             cq.assembly.Assembly | cq.occ_impl.shapes.Compound | cq.occ_impl.shapes.Solid
         ),
-        material_tags: list[str] | None,
+        material_tags: list[str] | None = None,
         scale_factor: float = 1.0,
     ) -> int:
         """Loads the parts from CadQuery object into the model.
@@ -602,18 +613,36 @@ class CadToDagmc:
         """
 
         if isinstance(cadquery_object, cq.assembly.Assembly):
-            cadquery_object = cadquery_object.toCompound()
+            # look for materials in each part of the assembly
+            if material_tags is None:
+                material_tags = []
+                for child in cadquery_object.children:
+                    if child.material is not None and child.material.name is not None:
+                        material_tags.append(child.material.name)
+                    else:
+                        raise ValueError(
+                            f"Not all parts in the assembly have material tags assigned. "
+                            f"Missing material tag for child: {child}. "
+                            "Please assign material tags to all parts or provide material_tags argument when adding the assembly."
+                        )
 
-        if isinstance(cadquery_object, (cq.occ_impl.shapes.Compound, cq.occ_impl.shapes.Solid)):
-            iterable_solids = cadquery_object.Solids()
+            cadquery_compound = cadquery_object.toCompound()
         else:
-            iterable_solids = cadquery_object.val().Solids()
+            cadquery_compound = cadquery_object
+
+        if isinstance(cadquery_compound, (cq.occ_impl.shapes.Compound, cq.occ_impl.shapes.Solid)):
+            iterable_solids = cadquery_compound.Solids()
+        else:
+            iterable_solids = cadquery_compound.val().Solids()
 
         if scale_factor == 1.0:
             scaled_iterable_solids = iterable_solids
         else:
             scaled_iterable_solids = [part.scale(scale_factor) for part in iterable_solids]
 
+        # look for materials in the
+        # if material_tags is None:
+
         check_material_tags(material_tags, scaled_iterable_solids)
         if material_tags:
             self.material_tags = self.material_tags + material_tags
@@ -686,6 +715,7 @@ class CadToDagmc:
             assembly.add(part)
 
         if imprint:
+            print("Imprinting assembly for unstructured mesh generation")
             imprinted_assembly, _ = cq.occ_impl.assembly.imprint(assembly)
         else:
             imprinted_assembly = assembly
@@ -776,6 +806,7 @@ class CadToDagmc:
             assembly.add(part)
 
         if imprint:
+            print("Imprinting assembly for mesh generation")
             imprinted_assembly, _ = cq.occ_impl.assembly.imprint(assembly)
         else:
             imprinted_assembly = assembly
@@ -811,7 +842,6 @@ class CadToDagmc:
     def export_dagmc_h5m_file(
         self,
         filename: str = "dagmc.h5m",
-        meshing_backend: str = "cadquery",
         implicit_complement_material_tag: str | None = None,
         scale_factor: float = 1.0,
         imprint: bool = True,
@@ -821,8 +851,6 @@ class CadToDagmc:
 
         Args:
             filename: the filename to use for the saved DAGMC file.
-            meshing_backend: determines whether gmsh or cadquery's direct mesh method
-                is used for meshing. Options are 'gmsh' or 'cadquery'.
             implicit_complement_material_tag: the name of the material tag to use
                 for the implicit complement (void space).
             scale_factor: a scaling factor to apply to the geometry.
@@ -830,6 +858,11 @@ class CadToDagmc:
 
             **kwargs: Backend-specific parameters:
 
+                Backend selection:
+                - meshing_backend (str, optional): explicitly specify 'gmsh' or 'cadquery'.
+                  If not provided, backend is auto-selected based on other arguments.
+                  Defaults to 'cadquery' if no backend-specific arguments are given.
+
                 For GMSH backend:
                 - min_mesh_size (float): minimum mesh element size
                 - max_mesh_size (float): maximum mesh element size
@@ -850,6 +883,52 @@ class CadToDagmc:
             ValueError: If invalid parameter combinations are used.
         """
 
+        # Define all acceptable kwargs
+        cadquery_keys = {"tolerance", "angular_tolerance"}
+        gmsh_keys = {
+            "min_mesh_size",
+            "max_mesh_size",
+            "mesh_algorithm",
+            "set_size",
+            "umesh_filename",
+            "method",
+            "unstructured_volumes",
+        }
+        all_acceptable_keys = cadquery_keys | gmsh_keys | {"meshing_backend"}
+
+        # Check for invalid kwargs
+        invalid_keys = set(kwargs.keys()) - all_acceptable_keys
+        if invalid_keys:
+            raise ValueError(
+                f"Invalid keyword arguments: {sorted(invalid_keys)}\n"
+                f"Acceptable arguments are: {sorted(all_acceptable_keys)}"
+            )
+
+        # Handle meshing_backend - either from kwargs or auto-detect
+        meshing_backend = kwargs.pop("meshing_backend", None)
+
+        if meshing_backend is None:
+            # Auto-select meshing_backend based on kwargs
+            has_cadquery = any(key in kwargs for key in cadquery_keys)
+            has_gmsh = any(key in kwargs for key in gmsh_keys)
+            if has_cadquery and not has_gmsh:
+                meshing_backend = "cadquery"
+            elif has_gmsh and not has_cadquery:
+                meshing_backend = "gmsh"
+            elif has_cadquery and has_gmsh:
+                provided_cadquery = [key for key in cadquery_keys if key in kwargs]
+                provided_gmsh = [key for key in gmsh_keys if key in kwargs]
+                raise ValueError(
+                    "Ambiguous backend: both CadQuery and GMSH-specific arguments provided.\n"
+                    f"CadQuery-specific arguments: {sorted(cadquery_keys)}\n"
+                    f"GMSH-specific arguments: {sorted(gmsh_keys)}\n"
+                    f"Provided CadQuery arguments: {provided_cadquery}\n"
+                    f"Provided GMSH arguments: {provided_gmsh}\n"
+                    "Please provide only one backend's arguments."
+                )
+            else:
+                meshing_backend = "cadquery"  # default
+
         # Validate meshing backend
         if meshing_backend not in ["gmsh", "cadquery"]:
             raise ValueError(
@@ -857,6 +936,8 @@ class CadToDagmc:
                 'Available options are "gmsh" or "cadquery"'
             )
 
+        print(f"Using meshing backend: {meshing_backend}")
+
         # Initialize variables to avoid unbound errors
         tolerance = 0.1
         angular_tolerance = 0.1
@@ -931,11 +1012,11 @@ class CadToDagmc:
 
         # Use the CadQuery direct mesh plugin
         if meshing_backend == "cadquery":
-            import cadquery_direct_mesh_plugin
 
             # Mesh the assembly using CadQuery's direct-mesh plugin
-            cq_mesh = assembly.toMesh(
-                imprint,
+            cq_mesh = to_mesh(
+                assembly,
+                imprint=imprint,
                 tolerance=tolerance,
                 angular_tolerance=angular_tolerance,
                 scale_factor=scale_factor,
@@ -962,6 +1043,7 @@ class CadToDagmc:
         elif meshing_backend == "gmsh":
             # If assembly is not to be imprinted, pass through the assembly as-is
             if imprint:
+                print("Imprinting assembly for mesh generation")
                 imprinted_assembly, imprinted_solids_with_org_id = cq.occ_impl.assembly.imprint(
                     assembly
                 )
